// Kernel for a single cell of a bias layer.

// Block size is usually [16 N] where N is found by experiment.  See CNS manual.
// N might need to be reduced for cards below GTX 285/295.
#BLOCKSIZE 16 16

// Get handle to the next level.  This allows us to use the 'HANDLE' versions of macros, which are somewhat faster.
COMPUTED_SENS_HANDLE hn = GET_COMPUTED_SENS_HANDLE(ZN);

// Get spatial dimensions of the next level.
int ySize = COMPUTED_SENS_HANDLE_Y_SIZE(hn);
int xSize = COMPUTED_SENS_HANDLE_X_SIZE(hn);
int dSize = COMPUTED_SENS_HANDLE_D_SIZE(hn);

// Average over the entire next level (for this feature).

float e = 0.0f;

for (int d = 0; d < dSize; d++) {
for (int x = 0; x < xSize; x++) {
for (int y = 0; y < ySize; y++) {

    e += READ_COMPUTED_SENS_HANDLE(hn, THIS_NF, y, x, d);

}
}
}

e /= (float)(ySize * xSize * dSize);

// Perform the bias update.

float b = READ_VAL;
b += ETA * e;
WRITE_VAL(b);
