// Kernel for a single cell of a hidden or output layer (forward pass).

// Block size is usually [16 N] where N is found by experiment.  See CNS manual.
// N might need to be reduced for cards below GTX 285/295.
#BLOCKSIZE 16 16

// Get handles to the previous level and weight layer.  This allows us to use the 'HANDLE' versions of macros,
// which are somewhat faster.
VAL_HANDLE        hp = GET_LEVEL_VAL_HANDLE (ZP);
WEIGHT_VAL_HANDLE hw = GET_WEIGHT_VAL_HANDLE(ZW);

// Get dimensions of weight array.
int fSize = WEIGHT_VAL_HANDLE_F_SIZE(hw);
int ySize = WEIGHT_VAL_HANDLE_Y_SIZE(hw);
int xSize = WEIGHT_VAL_HANDLE_X_SIZE(hw);
int dSize = WEIGHT_VAL_HANDLE_D_SIZE(hw);

// Find our receptive field in the previous level.
int y1, x1, d1, dummy;
FIND_LEVEL_Y_NEAREST(ZP, ySize, y1, dummy);
FIND_LEVEL_X_NEAREST(ZP, xSize, x1, dummy);
FIND_LEVEL_D_NEAREST(ZP, dSize, d1, dummy);

// Compute the dot product of the receptive field and the weight array (in 4-D).

float v = 0.0f;

for (int f = 0; f < fSize; f++) {
    for (int k = dSize - 1, d = d1; k >= 0; k--, d++) {  // We use the convolution convention for the spatial
    for (int j = xSize - 1, x = x1; j >= 0; j--, x++) {  // dimensions, flipping the weight array.
    for (int i = ySize - 1, y = y1; i >= 0; i--, y++) {

        float p = READ_VAL_HANDLE       (hp, f, y, x, d);
        float w = READ_WEIGHT_VAL_HANDLE(hw, f, i, j, k, THIS_F);

        v += p * w;

    }
    }
    }
}

// Add bias, apply nonlinearity, and save.

v += READ_BIAS_VAL(ZB, THIS_F);
v = Sigmoid(v);
WRITE_VAL(v);
