// Backward pass for a single cell of a hidden layer.  Computes the sensitivity.

// Block size is usually [16 N] where N is found by experiment.  See CNS manual.
// N might need to be reduced for cards below GTX 285/295.
#BLOCKSIZE 16 16

// Get handles to the next level and weight layer.  This allows us to use the 'HANDLE' versions of macros, which
// are somewhat faster.
SENS_HANDLE       hn = GET_COMPUTED_SENS_HANDLE(ZN );
WEIGHT_VAL_HANDLE hw = GET_WEIGHT_VAL_HANDLE   (ZNW);

// Get dimensions of weight array.
int ySize  = WEIGHT_VAL_HANDLE_Y_SIZE (hw);
int xSize  = WEIGHT_VAL_HANDLE_X_SIZE (hw);
int dSize  = WEIGHT_VAL_HANDLE_D_SIZE (hw);
int nfSize = WEIGHT_VAL_HANDLE_NF_SIZE(hw);

// Find our projective field in the next level.  Note that we are doing a 'full' correlation here, so we need to
// know the full projective field, as well as its valid portion.  In the case they are different, some part of the
// weight array will not be used.
int vy1, vy2, y1, dummy;
int vx1, vx2, x1;
int vd1, vd2, d1;
FIND_LEVEL_Y_NEAREST(ZN, ySize, vy1, vy2, y1, dummy);
FIND_LEVEL_X_NEAREST(ZN, xSize, vx1, vx2, x1, dummy);
FIND_LEVEL_D_NEAREST(ZN, dSize, vd1, vd2, d1, dummy);

// Compute the dot product of the projective field's sensitivity map and the weight array (in 4-D).

float s = 0.0f;

for (int nf = 0; nf < nfSize; nf++) {
    for (int d = vd1, k = vd1 - d1; d <= vd2; d++, k++) {  // Because the forward pass used the convolution
    for (int x = vx1, j = vx1 - x1; x <= vx2; x++, j++) {  // convention, we have to use the correlation
    for (int y = vy1, i = vy1 - y1; y <= vy2; y++, i++) {  // convention here.  We don't flip the weight array.

        float n = READ_SENS_HANDLE      (hn, nf    , y, x, d);
        float w = READ_WEIGHT_VAL_HANDLE(hw, THIS_F, i, j, k, nf);

        s += n * w;

    }
    }
    }
}

// Multiply by the derivative of the forward-propagating value and store.

s *= DSigmoid(READ_VAL);
WRITE_SENS(s);
