// Kernel for a single cell of a weight layer.

// Block size is usually [16 N] where N is found by experiment.  See CNS manual.
// N might need to be reduced for cards below GTX 285/295.
#BLOCKSIZE 16 16

// Get handles to the previous and next levels.  This allows us to use the 'HANDLE' versions of macros, which are
// somewhat faster.
LEVEL_VAL_HANDLE     hp = GET_LEVEL_VAL_HANDLE    (ZP);
COMPUTED_SENS_HANDLE hn = GET_COMPUTED_SENS_HANDLE(ZN);

// Get spatial dimensions of the next level.  It will always be smaller than the previous level.
int nySize = COMPUTED_SENS_HANDLE_Y_SIZE(hn);
int nxSize = COMPUTED_SENS_HANDLE_X_SIZE(hn);
int ndSize = COMPUTED_SENS_HANDLE_D_SIZE(hn);

// Find the corresponding subcube in the previous level.  This depends on which weight cell we are updating.
// Note we are performing a valid correlation.
int py1 = Y_SIZE - 1 - THIS_Y;
int px1 = X_SIZE - 1 - THIS_X;
int pd1 = D_SIZE - 1 - THIS_D;

// Compute the dot product of the next level and its corresponding subcube in the previous level (in 3-D).
// Normalize by number of elements.

float e = 0.0f;

for (int nd = 0, pd = pd1; nd < ndSize; nd++, pd++) {  // This is a correlation.  Nothing gets flipped.
for (int nx = 0, px = px1; nx < nxSize; nx++, px++) {
for (int ny = 0, py = py1; ny < nySize; ny++, py++) {

    float p = READ_LEVEL_VAL_HANDLE    (hp, THIS_F , py, px, pd);
    float n = READ_COMPUTED_SENS_HANDLE(hn, THIS_NF, ny, nx, nd);

    e += p * n;

}
}
}

e /= (float)(nySize * nxSize * ndSize);

// Perform the weight update.

float w = READ_VAL;
w += ETA * e;
WRITE_VAL(w);
